#!/usr/bin/python

import sys
import re
import pysam
import os
from collections import Counter
from contextlib import contextmanager

bamfile = sys.argv[1]
outfolder = sys.argv[2]
barcodeTag = sys.argv[3]
min_barcodes = int(sys.argv[4])
mtchr = sys.argv[5]
quant_file = sys.argv[6]
passing_file = sys.argv[7]

base=os.path.basename(bamfile)
basename=os.path.splitext(base)[0]

def getBarcode(intags):
	'''
	Parse out the barcode per-read
	'''
	for tg in intags:
		if(barcodeTag == tg[0]):
			return(tg[1])
	return("NA")

def quantifyBarcodes(mtchr):
	'''
	Make a giant dictionary of observed barcodes at the mitochondrial chr
	'''
	barcodes_all = dict()
	bam = pysam.AlignmentFile(bamfile,'rb')
	Itr = bam.fetch(str(mtchr),multiple_iterators=False)
	
	for read in Itr:
		read_barcode = getBarcode(read.tags)
		barcodes_all[read_barcode] = barcodes_all.get(read_barcode, 0) + 1
	bam.close()
	return(barcodes_all)


# Quant barcodes and write it out
barcodes = quantifyBarcodes(mtchr)
barcodes = {x : barcodes[x] for x in barcodes if barcodes[x] >= min_barcodes and x != "NA"}
bc = list(barcodes.keys())	
	
quant_file_o = open(quant_file, "w") 
for k, v in barcodes.items():
	quant_file_o.write(k +","+ str(v)+"\n")
quant_file_o.close() 

passing_file_o = open(passing_file, "w") 
for k, v in barcodes.items():
	passing_file_o.write(k +"\n")
passing_file_o.close() 



#!/usr/bin/python

import sys
import re
import pysam
import os
from collections import Counter

bamfile = sys.argv[1]
outfolder = sys.argv[2]
barcodeTag = sys.argv[3]
bcfile = sys.argv[4]
mtchr = sys.argv[5]
umitag = sys.argv[6]

base=os.path.basename(bamfile)
basename=os.path.basename(os.path.splitext(bcfile)[0])

def getBarcode(read, tag_get):
	'''
	Parse out the barcode per-read
	'''
	# for tg in read.tags:
	# 	if(tag_get == tg[0]):
	# 		return(tg[1])
	# return("AA")
	# Using get_tag to get the results
	try:
		return read.get_tag(tag_get)
	except Exception as e:
		return ("AA")


# Read in the barcodes
with open(bcfile) as barcode_file_handle:
    content = barcode_file_handle.readlines()
bc = set([x.strip() for x in content]) # from list to set, so query becomes O(1)

bam = pysam.AlignmentFile(bamfile, "rb")
outname = outfolder + "/" + basename + ".bam"
out = pysam.AlignmentFile(outname, "wb", template = bam)

# Make a DNA inspired additional barcode to account for potentially different sample indices
bases = "ACGT"
fauxdon = [a + b + c + d for a in bases for b in bases for c in bases for d in bases]

# Filter for reads that match the set of possible barcodes for this sample
try:
	Itr = bam.fetch(str(mtchr),multiple_iterators=False)
	for read in Itr:
		barcode_id = getBarcode(read, barcodeTag)
		
		if(barcode_id in bc):
		
			# Now check for true UMI
			if(umitag != "XX"): 
				umi_id = getBarcode(read, umitag)
			else:
				umi_id = ""
			
			# Make a fake UMI from 1) cell barcode + 2) captured umi + 3) experiment
			# all with just ACGTs so that picard doesn't bark at us. 
			# Only do this if the last string element is a number (i.e channel in 10x convention)
			if(barcode_id[-1].isnumeric()):
				split_two = barcode_id.split("-")
				faux_umi = split_two[0] + umi_id + fauxdon[(int(split_two[1]) - 1)]
			else:
				faux_umi = barcode_id + umi_id 
			read.tags = read.tags + [("MU", faux_umi)]
			out.write(read)
			

except OSError: # Truncated bam file from previous iteration handle
	print('Finished parsing bam')
	
bam.close()
out.close()

pysam.index(outname)




#!/usr/bin/python

###################################################
# Summarizes the total number of reads per position / strand
###################################################

import sys
import re
import os
import pysam
import numpy as np

bam_file = sys.argv[1]
barcodes_file = sys.argv[2]
out_pre = sys.argv[3]
max_bp = int(sys.argv[4])
base_qual = float(sys.argv[5])
fasta_file = sys.argv[6]
alignment_quality = float(sys.argv[7])
barcode_tag = sys.argv[8]

# Import barcodes
with open(barcodes_file) as barcode_file_handle:
    content = barcode_file_handle.readlines()
bcs = [x.strip() for x in content]
bam_input = pysam.AlignmentFile(bam_file, "rb")
dna_letters = ['A','C','G','T']

def getBarcode(intags):
	'''
	Parse out the barcode per-read
	'''
	for tg in intags:
		if(barcode_tag == tg[0]):
			return(tg[1])
	return("NA")


# Dimension cell x position x letter x strand
# Coverage associated with the bases
ca =  np.zeros((len(bcs),max_bp,4,2), dtype=int)

for read in bam_input:
	
	if(read.is_reverse):
		s_idx = 1
	else:
		s_idx = 0
		
	# Get read attributes 
	seq = read.seq
	quality = read.query_qualities
	align_qual_read = read.mapping_quality
	cell_barcode = getBarcode(read.tags)
	if(cell_barcode != "NA"):
		c_idx = bcs.index(cell_barcode)
	
		for q_idx, p_idx in read.get_aligned_pairs(True):
			if q_idx is not None and p_idx is not None and align_qual_read > alignment_quality:
				if(quality[q_idx] > base_qual and seq[q_idx] in dna_letters):
					l_idx = dna_letters.index(seq[q_idx])
					ca[c_idx,p_idx,l_idx,s_idx] += 1


# Function to write the slice of the matrix that is associated with the 
def writeSparseMatrixLetter(letter, letter_idx):
	out_file_fn = out_pre + "."+letter+".txt"
	with open(out_file_fn,"w") as file_handle_fn:
		for cell_idx, cell_name in enumerate(bcs):
			
			# Pull out the stranded counts
			fw_vec = ca[cell_idx,:,letter_idx,0].ravel()
			rev_vec = ca[cell_idx,:,letter_idx,1].ravel()
			
			# Write each position
			for i in range(0,int(max_bp)):
				if(fw_vec[i] > 0 or rev_vec[i] > 0):
					file_handle_fn.write(str(i+1)+","+cell_name+","+str(fw_vec[i])+","+str(rev_vec[i])+"\n")

writeSparseMatrixLetter("A", 0)
writeSparseMatrixLetter("C", 1)
writeSparseMatrixLetter("G", 2)
writeSparseMatrixLetter("T", 3)

# Export the per-base coverage for the thrill of it and the depth
out_file_depth = out_pre.replace("/temp/sparse_matrices/", "/qc/depth/") + ".depth.txt"
out_file_coverage= out_pre + ".coverage.txt"
with open(out_file_coverage,"w") as file_handle_cov:
	with open(out_file_depth,"w") as file_handle_depth:
	
		# Loop over cells
		for cell_idx, cell_name in enumerate(bcs):
		
			# Pull out the summed counts per cell per position
			cov_vec = np.sum(ca[cell_idx,:,:,:], axis = (1,2)).tolist()
			depth = round(sum(cov_vec)/len(cov_vec),2)
			# Write each position
			for i in range(0,int(max_bp)):
				if(cov_vec[i] > 0):
					file_handle_cov.write(str(i+1)+","+cell_name+","+str(cov_vec[i])+"\n")
			
			# Now write the depth
			file_handle_depth.write(cell_name+"\t"+str(depth)+"\n")







#!/usr/bin/python

import sys
import re
import pysam
import os
from collections import Counter
from contextlib import contextmanager

bamfile = sys.argv[1]
outfolder = sys.argv[2]
barcodeTag = sys.argv[3]
bcfile = sys.argv[4]
mtchr = sys.argv[5]

base=os.path.basename(bamfile)
basename=os.path.splitext(base)[0]

def getBarcode(read):
	'''
	Parse out the barcode per-read
	'''
	if read.has_tag(barcodeTag):
		return read.get_tag(barcodeTag)
	else:
		return "NA"


def writePassingReads(bc_dict, mtchr):
	'''
	Write out reads to their corresponding files based on a barcode index
	'''
	bam = pysam.AlignmentFile(bamfile,'rb')
	Itr = bam.fetch(str(mtchr),multiple_iterators=False)
	for read in Itr:
		read_barcode = getBarcode(read)
		
		# If read barcode is in whitelist, then write it out
		if read_barcode in bc_dict:
			idx = bc_dict[read_barcode]
			file = fopen[idx]
			file.write(read)

# Read in the barcodes
with open(bcfile) as barcode_file_handle:
    content = barcode_file_handle.readlines()
bc = [x.strip() for x in content]

# Open up a bunch of files and write out reads for valid barcodes
@contextmanager
def multi_file_manager(files, mode='rt'):
	"""
	Open multiple files and make sure they all get closed.
	"""
	temp = pysam.AlignmentFile(bamfile, "rb")
	files = [pysam.AlignmentFile(file, "wb", template = temp) for file in files]
	temp.close()
	yield files
	for file in files:
		file.close()
		
# Final loop to write out passing reads
bambcfiles = [outfolder + "/" + bc1 + ".bam" for bc1 in bc]
bc_dict = {bc1: i  for i,bc1 in enumerate(bc)}
with multi_file_manager(bambcfiles) as fopen:
	writePassingReads(bc_dict, mtchr)






import click
import os
import os.path
import sys
import shutil
import random
import string
import time
import pysam
import numpy as np
import re
import csv

from pkg_resources import get_distribution


@click.command()
@click.version_option()
@click.option('--input', '-i', default = ".", required=True, help='Input; a single .bam file of reads to be processed.')
@click.option('--mito-chromosome', '-mc', default = "chrM", required=True, help='Name of mtDNA chromosome in bam file (e.g. chrM or MT)')
@click.option('--output', '-o', default = "mgatkdel_find", required=True, help='Name of output files prefix')

def main(input, mito_chromosome, output):
	
	"""
	mgatk-del-find: detect possible deletion junctions from bam files. \n
	See: `mgatk-del-find --help`
	"""
	
	
	def gettime(): 
		"""
		Matches `date` in Linux
		"""
		return(time.strftime("%a ") + time.strftime("%b ") + time.strftime("%d ") + time.strftime("%X ") + 
			time.strftime("%Z ") + time.strftime("%Y")+ ": ")

	
	script_dir = os.path.dirname(os.path.realpath(__file__))
	cwd = os.getcwd()
	__version__ = get_distribution('mgatk').version
	click.echo(gettime() + "mgatk-del-find v%s" % __version__)
	R_plot_script = script_dir + "/bulk_del/plot_deletion_breaks_bulk.R"
	
	bam_in = pysam.AlignmentFile(input, "rb")
	click.echo(gettime() + "Processing .bam file.")
	
	def process_cigar_for_clip_position(cigar, tuple):

		pos = None

		# Case 1/2: start of read
		if(cigar[1] == "H" or cigar[2] == "H" or
			cigar[1] == "S" or cigar[2] == "S"):
			pos = tuple[0][1]

		# Case 2: end of read
		if(cigar[-1] == "H" or cigar[-1] == "S"):
			pos = tuple[-1][1]

		return(pos)
		
	# function to get tags
	def getTag(intags, tag):
		for tg in intags:
			if(tag == tg[0]):
				return(tg[1])
		return(None)
		
	def SA_pos(SA_tag):
		position = int(SA_tag.split(',')[1]) - 1 # 0-based
		if SA_tag.split(',')[3][1] == "M" or SA_tag.split(',')[3][2] == "M":
			position += (int(SA_tag.split(',')[3].split('M')[0]) - 1)
		return(position)
		
	# iniate empty lists for storing counts
	# loop and extract per position clipping counts

	clip_pos_count_0 = [0] * 16569
	SA_count_0 = [0] * 16569
	clip_pos_count = [0] * 16569
	SA_count = [0] * 16569
	out1_list = []
	out2_list = []
	
	for read in bam_in.fetch(mito_chromosome):
		seq = read.seq
		reverse = read.is_reverse
		cigar_string = read.cigarstring
		positions = read.get_reference_positions()
		tuple = read.get_aligned_pairs(True)
		if positions and cigar_string:
			SA_tag = getTag(read.tags, 'SA')
			clip_pos = process_cigar_for_clip_position(cigar_string, tuple)
			start_end = read.get_reference_positions()
			if clip_pos is not None:
				clip_pos_count[clip_pos] += 1
			if SA_tag is not None:
				if str(SA_tag.split(',')[0]) == str(mito_chromosome):
					position = SA_pos(SA_tag)
					SA_count[position] += 1
					if SA_tag.split(',')[3][1] == 'M' or SA_tag.split(',')[3][2] == 'M':
						out2_list.append(int(start_end[0] + 1))
					else:
						out2_list.append(int(start_end[-1] + 1))
					out1 = SA_pos(SA_tag) + 1
					out1_list.append(out1)

	clip_pos_count = np.array(clip_pos_count)
	SA_count = np.array(SA_count)

	cov = bam_in.count_coverage(mito_chromosome, quality_threshold=0, read_callback='nofilter')
	cov_out = np.array(np.add(np.add(cov[0], cov[1]), np.add(cov[2], cov[3])).tolist())

	outfile_clip = output + '.clip.tsv'
	with open(outfile_clip, 'w') as f:
		writer = csv.writer(f, delimiter='\t')
		idx = np.argsort(-clip_pos_count)
		writer.writerow(['position', 'coverage', 'clip_count', 'SA'])
		writer.writerows(zip(np.array(range(1, 16570))[idx], cov_out[idx], clip_pos_count[idx], SA_count[idx]))

	outputSA = output + '.SA.tsv'
	with open(outputSA, 'w') as f:
		writer = csv.writer(f, delimiter='\t')
		writer.writerow(['out1', 'out2'])
		writer.writerows(zip(out1_list, out2_list))
	
	# Make the R call
	click.echo(gettime() + "Visualizing results.")
	Rcall = "Rscript " + R_plot_script + " " + outfile_clip + " " + outputSA
	os.system(Rcall)
		


#!/usr/bin/python

###################################################
# Call mito variants
###################################################

import sys
import gzip
import glob
import numpy as np
import pandas as pd

import matplotlib
from matplotlib import pyplot as plt


def load_mgatk_output(output_dir, mito_length):
	# assuming mgatk output naming convention
	base_files = [glob.glob(output_dir + '*.{}.txt.gz'.format(nt))[0] for nt in 'ATCG']
	
	base_coverage_dict = dict()
	for i, nt in enumerate('ATCG'):
		cur_base_data = pd.read_csv(gzip.open(base_files[i]), header=None)
		
		# gather coverage for forward strand
		fwd_base_df = cur_base_data[[0, 1, 2]].pivot_table(index=1, columns=0)
		fwd_base_df.columns = [x[1] for x in fwd_base_df.columns.values]  # flatten weird multiindex after pivot
		fwd_base_df.index.name = None
		missing_pos = [x for x in range(1, mito_length+1) if x not in fwd_base_df.columns]
		fwd_base_df[missing_pos] = 0  # fill in missing positions
		fwd_base_df = fwd_base_df.fillna(0).sort_index(axis=1)  # assume all nan are true zeroes
		
		# gather coverage for forward strand
		rev_base_df = cur_base_data[[0, 1, 3]].pivot_table(index=1, columns=0)
		rev_base_df.columns = [x[1] for x in rev_base_df.columns.values]
		rev_base_df.index.name = None
		missing_pos = [x for x in range(1, mito_length+1) if x not in rev_base_df.columns]
		rev_base_df[missing_pos] = 0
		rev_base_df = rev_base_df.fillna(0).sort_index(axis=1)
		
		# organize base data into a dict
		base_coverage_dict[nt] = (fwd_base_df, rev_base_df)
	
	return base_coverage_dict


def gather_possible_variants(base_coverage_dict, reference_file):
	# sum across cells and strands for each base and position
	aggregated_genotype = pd.DataFrame(np.zeros((4, mito_length)), index=list('ATCG'), columns=np.arange(1, mito_length+1))
	for nt in base_coverage_dict:
		# sum across cells for each strand separately
		fwd_base_df, rev_base_df = base_coverage_dict[nt]
		fwd_base_sum, rev_base_sum = fwd_base_df.sum(), rev_base_df.sum()
		
		# sequencing artifact if a base/position is only nonzero for one strand across cells, ignore them
		masking = ~((fwd_base_sum > 0) & (rev_base_sum > 0))  # True if position not >0 for both strands
		fwd_base_sum[masking], rev_base_sum[masking] = 0, 0
		
		# sum across strands
		aggregated_genotype.loc[nt, :] = fwd_base_sum + rev_base_sum
	
	# make a reference set of tuples (pos, ref_base)
	ref_set = [x.strip().split() for x in open(reference_file, 'r').readlines()]
	ref_N_positions = [int(x[0]) for x in ref_set if x[1].upper() not in letters]
	ref_set = set([(int(x[0]), x[1].upper()) for x in ref_set if x[1].upper() in letters])
	ref_dict = dict(ref_set)
	
	# make an observed set of tuples which are nonzero
	non_zero_idx = np.where(aggregated_genotype > 0)
	non_zero_bases = [letters[i] for i in non_zero_idx[0]]
	non_zero_pos = [int(i+1) for i in non_zero_idx[1]]
	observed_set = list(zip(non_zero_pos, non_zero_bases))
	observed_set = set([x for x in observed_set if x[0] not in ref_N_positions])  # disregard positions in ref with N
	
	# take difference between observed and reference
	variant_set = observed_set - ref_set
	variants = sorted([(x[0], ref_dict[x[0]], x[1]) for x in list(variant_set)], key=lambda x: x[0])  # (pos, ref base, obs base)
	
	return variants

# Parse arguments
MGATK_OUT_DIR = sys.argv[1]
sample_prefix = sys.argv[2]
mito_length = int(sys.argv[3])  # 16569
low_coverage_threshold = int(sys.argv[4])  # 10
mito_genome = sys.argv[5]  # chrM

letters = list('ATCG')

base_coverage_dict = load_mgatk_output(MGATK_OUT_DIR, mito_length)
cell_barcodes = base_coverage_dict['A'][0].index

# total coverage per position per cell
total_coverage = pd.DataFrame(np.zeros((len(cell_barcodes), mito_length)), index=cell_barcodes, columns=np.arange(1, mito_length+1))
for nt in base_coverage_dict:
	total_coverage += base_coverage_dict[nt][0]
	total_coverage += base_coverage_dict[nt][1]

# exclude low coverage cells from variant calling
cell_barcodes = total_coverage.index[total_coverage.mean(axis=1) > low_coverage_threshold]
for nt in base_coverage_dict:
    base_coverage_dict[nt] = (base_coverage_dict[nt][0].loc[cell_barcodes, :], base_coverage_dict[nt][1].loc[cell_barcodes, :])
total_coverage = total_coverage.loc[cell_barcodes, :]

# call potential variants
variants = gather_possible_variants(base_coverage_dict, MGATK_OUT_DIR + mito_genome +'_refAllele.txt')
variant_names = ['{}{}>{}'.format(x[0], x[1], x[2]) for x in variants]

# build two <cell by variant tables>, one for each strand
total_coverage_variant_df = []
fwd_cell_variant_df, rev_cell_variant_df = [], []
for i, var in enumerate(variants):
    var_name = variant_names[i]
    pos, base = var[0], var[2]
    total_coverage_variant_df.append(total_coverage[pos])
    fwd_cell_variant_df.append(base_coverage_dict[base][0][pos].values)
    rev_cell_variant_df.append(base_coverage_dict[base][1][pos].values)
total_coverage_variant_df = pd.DataFrame(np.array(total_coverage_variant_df).T, index=cell_barcodes, columns=variant_names)
fwd_cell_variant_df = pd.DataFrame(np.array(fwd_cell_variant_df).T, index=cell_barcodes, columns=variant_names)
rev_cell_variant_df = pd.DataFrame(np.array(rev_cell_variant_df).T, index=cell_barcodes, columns=variant_names)
all_cell_variant_df = fwd_cell_variant_df + rev_cell_variant_df

# heteroplasmic ratio
heteroplasmic_df = all_cell_variant_df / total_coverage_variant_df

# strand correlation
mask_idx = (fwd_cell_variant_df + rev_cell_variant_df) == 0  # set 0 on both strands to nan to exclude from correlation calculation
fwd_cell_variant_df[mask_idx] = np.nan
rev_cell_variant_df[mask_idx] = np.nan
variant_strand_corr = fwd_cell_variant_df.corrwith(rev_cell_variant_df).round(3)

# vmr
variant_mean = all_cell_variant_df.sum() / total_coverage_variant_df.sum()
variant_var = heteroplasmic_df.var()
variant_vmr = variant_var / (variant_mean + 0.00000000001)

# compute other summary stats
variant_positon = [x[0] for x in variants]
variant_nucleotide = ['{}>{}'.format(x[1], x[2]) for x in variants]
variant_n_cells_conf_detected = ((fwd_cell_variant_df >= 2) & (rev_cell_variant_df >= 2)).sum()
variant_n_cells_over_5 = (heteroplasmic_df >= 0.05).sum()
variant_n_cells_over_10 = (heteroplasmic_df >= 0.1).sum()
variant_n_cells_over_20 = (heteroplasmic_df >= 0.2).sum()
variant_n_cells_over_95 = (heteroplasmic_df >= 0.95).sum()
max_heteroplasmy = heteroplasmic_df.max()
variant_mean_coverage = total_coverage_variant_df.mean()

# pack summary stats
variant_output = pd.DataFrame([variant_positon, variant_nucleotide, variant_names,
                            variant_vmr, variant_mean, variant_var,
                            variant_n_cells_conf_detected, variant_n_cells_over_5,
                            variant_n_cells_over_10, variant_n_cells_over_20, variant_n_cells_over_95,
                            max_heteroplasmy, variant_strand_corr, variant_mean_coverage]).T
variant_output.columns = ['position', 'nucleotide', 'variant', 'vmr', 'mean', 'variance',
                          'n_cells_conf_detected', 'n_cells_over_5',
                          'n_cells_over_10', 'n_cells_over_20', 'n_cells_over_95',
                          'max_heteroplasmy', 'strand_correlation', 'mean_coverage']
variant_output[['vmr', 'mean', 'variance', 'strand_correlation', 'mean_coverage', 'max_heteroplasmy']] = variant_output[['vmr', 'mean', 'variance', 'strand_correlation',
                                                                                                                         'mean_coverage', 'max_heteroplasmy']].astype(np.float)

# exclude variants with less than three cells
multi_cell_variants = variant_output[variant_output['n_cells_conf_detected'] >= 3]['variant']
heteroplasmic_df = heteroplasmic_df[multi_cell_variants]

# generate caleb plot
plt.figure(figsize=(10, 8))
plt.scatter(variant_output[variant_output['variant'].isin(multi_cell_variants)]['strand_correlation'],
	np.log10(variant_output[variant_output['variant'].isin(multi_cell_variants)]['vmr']), s=5)
plt.axhline(np.log10(0.01), color='red', alpha=0.4, linestyle=':')
plt.axvline(0.65, color='red', alpha=0.4, linestyle=':')
plt.xlabel('strand correlation', fontsize=20)
plt.ylabel('log10(VMR)', fontsize=20)

# save results
plt.savefig(MGATK_OUT_DIR + sample_prefix + '.vmr_strand_plot.png')
variant_output.to_csv(MGATK_OUT_DIR + sample_prefix + '.variant_stats.tsv.gz', sep='\t',compression='gzip', index=False)
heteroplasmic_df.to_csv(MGATK_OUT_DIR + sample_prefix + '.cell_heteroplasmic_df.tsv.gz', sep='\t',compression='gzip')



#!/usr/bin/python

###############################################
# Filter reads / clip .bam files
# Don't print() anything!!!! writing to STDOUT
##############################################

import sys
import re
import pysam

bamfile = sys.argv[1]
logfile = sys.argv[2]
mtchr = sys.argv[3]
proper_pair = sys.argv[4]
NHmax = sys.argv[5]
NMmax = sys.argv[6]

# https://github.com/pysam-developers/pysam/issues/509
# https://pysam.readthedocs.io/en/latest/api.html using text
bam = pysam.AlignmentFile(bamfile, "rb")

# Modify the header to account for CRA v2 nonsense
new_header = str(bam.header).replace("@HD\tSO:coordinate", "@HD\tVN:1.5\tSO:coordinate")
out = pysam.AlignmentFile("-", "wb", text = new_header)

keepCount = 0
filtCount = 0

def filterReadTags(intags):
    '''
    Checks for aligner-specific read tags and filters
	'''
	
    for tg in intags:
    	if(('NH' == tg[0] and int(tg[1]) > int(NHmax)) or \
    		(('NM' == tg[0] or 'nM' == tg[0]) and int(tg[1]) > int(NMmax))):
        		return(False)
    return(True)

def pairing(read):
	'''
	Check if read is paired, properly paired, etc.
	'''
	
	if(proper_pair != "True"): # then user doesn't care to filter it
		return(True)
	else:
		return(read.is_proper_pair)

def processRead(read):
	global keepCount
	global filtCount
	if(filterReadTags(read.tags) and read.reference_name == mtchr and pairing(read)):
		keepCount += 1
		out.write(read)
	else:
		filtCount += 1

for read in bam:
	processRead(read)

with open(logfile , 'w') as outfile:
	outfile.write("Kept "+ str(keepCount) + "\n" + "Removed " + str(filtCount)+ "\n")




#!/usr/bin/python

from os.path import join
import os
import subprocess
import sys
import shutil
import pysam
from ruamel import yaml

configFile = sys.argv[1]
inputbam = sys.argv[2]
outputbam = sys.argv[3]
sample = sys.argv[4]

with open(configFile, 'r') as stream:
	config = yaml.load(stream, Loader=yaml.Loader)

# Parse the configuration variables
indir = config["input_directory"]
outdir = config["output_directory"]
script_dir = config["script_dir"]

mito_genome = config["mito_chr"]
mito_length = str(config["mito_length"])
fasta_file = config["fasta_file"]

remove_duplicates = config["remove_duplicates"]
umi_barcode = config["umi_barcode"]
emit_base_qualities = config["emit_base_qualities"]

handle_overlap = config["handle_overlap"]
proper_paired = config["proper_paired"]
base_qual = str(config["base_qual"])
alignment_quality = config["alignment_quality"]
NHmax = config["NHmax"]
NMmax = config["NMmax"]

max_javamem  = config["max_javamem"]

# Software paths
java = "java"
python = "python"

# Script locations
filtclip_py = script_dir + "/bin/python/filterClipBam.py"
detailedcall_py = script_dir + "/bin/python/detailedCalls.py"
sumstatsBP_py = script_dir + "/bin/python/sumstatsBP.py"
sumstatsBP_overlap_py = script_dir + "/bin/python/sumstatsBP_overlap.py"
picardCall = java + " -Xmx"+max_javamem+"  -jar " + script_dir + "/bin/picard.jar MarkDuplicates"

# Prepare filepath locations
rmlog = outputbam.replace(".qc.bam", ".rmdups.log").replace("/temp/ready_bam/", "/logs/rmdupslogs/")
filtlog = outputbam.replace(".qc.bam", ".filter.log").replace("/temp/ready_bam/", "/logs/filterlogs/")
temp_bam0 = outputbam.replace(".qc.bam", ".temp0.bam").replace("/temp/ready_bam/", "/temp/temp_bam/")
temp_bam1 = outputbam.replace(".qc.bam", ".temp1.bam").replace("/temp/ready_bam/", "/temp/temp_bam/")
prefixSM = outdir + "/temp/sparse_matrices/" + sample
outputdepth = outdir + "/qc/depth/" + sample + ".depth.txt"

# 1) Filter bam files
pycall = " ".join([python, filtclip_py, inputbam, filtlog, mito_genome, proper_paired, NHmax, NMmax]) + " > " + temp_bam0
os.system(pycall)

# 2) Sort the filtered bam file
pysam.sort("-o", temp_bam1, temp_bam0)
pysam.index(temp_bam1)

# See if we have UMIs
if(umi_barcode != "" and len(umi_barcode) == 2):
	umi_extra = " BARCODE_TAG=" + umi_barcode
else:
	umi_extra = "" 

# 3) (Optional) Remove duplicates
if (remove_duplicates == "True"):
	mdc_long = picardCall + " I="+temp_bam1+" O="+outputbam+" M="+rmlog+" REMOVE_DUPLICATES=true VALIDATION_STRINGENCY=SILENT QUIET=true VERBOSITY=ERROR USE_JDK_DEFLATER=true ASSUME_SORTED=true USE_JDK_INFLATER=true" + umi_extra 
	proc = subprocess.Popen(mdc_long, stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
	out, err = proc.communicate()  # Read data from stdout and stderr
else: # just move the previous output
	os.system("mv " + temp_bam1 + " " + outputbam)
	os.system("rm " + temp_bam1 + ".bai")
pysam.index(outputbam)

# 4) Get allele counts per sample / base pair and per-base quality scores
if (handle_overlap == "True"):
	alleleCountcall = " ".join([python, sumstatsBP_overlap_py, outputbam, prefixSM, mito_genome, mito_length, base_qual, sample, fasta_file, alignment_quality, emit_base_qualities])
else:
	alleleCountcall = " ".join([python, sumstatsBP_py, outputbam, prefixSM, mito_genome, mito_length, base_qual, sample, fasta_file, alignment_quality, emit_base_qualities])
os.system(alleleCountcall)

# 5) Get depth from the coverage sparse matrix
with open(prefixSM + ".coverage.txt", 'r') as coverage:
	depth = 0
	for row in coverage:
		s = row.split(",")
		depth += int(s[2].strip())
with open(outputdepth, 'w') as d:
	d.write(sample + "\t" + str(round(float(depth)/float(mito_length),2)) + "\n")




import itertools
import time
import shutil
import re
import os
import sys
import subprocess
import pysam
import filecmp
import math
import pysam

def string_hamming_distance(str1, str2):
    """
    Fast hamming distance over 2 strings known to be of same length.
    In information theory, the Hamming distance between two strings of equal 
    length is the number of positions at which the corresponding symbols 
    are different.
    eg "karolin" and "kathrin" is 3.
    """
    return sum(itertools.imap(operator.ne, str1, str2))


def rev_comp(seq):
    """
    Fast Reverse Compliment
    """  
    tbl = {'A':'T', 'T':'A', 'C':'G', 'G':'C', 'N':'N'}
    return ''.join(tbl[s] for s in seq[::-1])

def gettime(): 
	"""
	Matches `date` in Linux
	"""
	return(time.strftime("%a ") + time.strftime("%b ") + time.strftime("%d ") + time.strftime("%X ") + 
		time.strftime("%Z ") + time.strftime("%Y")+ ": ")

def findIdx(list1, list2):
	"""
	Return the indices of list1 in list2
	"""
	return [i for i, x in enumerate(list1) if x in list2]

def check_R_packages(required_packages):
	"""
	Determines whether or not R packages are properly installed
	"""
	R_path = shutil.which("R")
	installed_packages = os.popen(R_path + ''' -e "installed.packages()" | awk '{print $1}' | sort | uniq''').read().strip().split("\n")
	if(not set(required_packages) < set(installed_packages)):
		sys.exit("ERROR: cannot find the following R package: " + str(set(required_packages) - set(installed_packages)) + "\n" + 
			"Install it in your R console and then try rerunning mgatk (but there may be other missing dependencies).")

def check_software_exists(tool):
	tool_path = shutil.which(tool)
	if(str(tool_path) == "None"):
		sys.exit("ERROR: cannot find "+tool+" in environment; add it to user PATH environment")


def parse_fasta(filename):
	"""
	Imports specified .fasta file
	"""
	f = open(filename)
	sequences = {}
	for line in f:
		if line.startswith('>'):
			name = line[1:].strip()
			sequences[name] = ''
		else:
			sequences[name] = sequences[name] + line.strip()
	f.close()
	return sequences

def verify_bai(bamfile):
	'''
	Function that indexes bam file from input if missing
	'''
	bai_file = bamfile + ".bai"
	if(not os.path.exists(bai_file)):
		pysam.index(bamfile)

# Helper function for chunking the bam
def split_chunk_file(one_barcode_file, script_dir, input, bcbd, barcode_tag, mito_chr, umi_barcode):
	chunk_bam_py = script_dir + "/bin/python/chunk_barcoded_bam.py"
	pycall = " ".join(['python', chunk_bam_py, input, bcbd, barcode_tag, one_barcode_file, mito_chr, umi_barcode])
	os.system(pycall)

def verify_sample_mitobam(bam, mito_chr, mito_length):
	idxs = pysam.idxstats(bam).split("\n")
	nReads = 0
	bam_length = 0
	# Pull out essentials from idxstats
	for i in idxs:
		if(i.split("\t")[0] == mito_chr):
			bam_length = int(i.split("\t")[1])
			nReads = int(i.split("\t")[2])
			
	# If missing for some reason and don't want to check
	if(mito_length == -9):
		mito_length = bam_length
	return(bam_length == mito_length and nReads > 0)


def handle_fasta_inference(mito_genome, supported_genomes, script_dir, mode, of, write_files = True):
	"""
	Determines what's going on with the mitochondrial genome
	based on user input / existing data
	"""
	if any(mito_genome in s for s in supported_genomes):
		fastaf = script_dir + "/bin/anno/fasta/" + mito_genome + ".fasta"
	else:
		if os.path.exists(mito_genome):
			fastaf = mito_genome
		else:
			sys.exit('ERROR: Could not find file ' + mito_genome + '; QUITTING')
	fasta = parse_fasta(fastaf)	

	if(len(fasta.keys()) != 1):
		sys.exit('ERROR: .fasta file has multiple chromosomes; supply file with only 1; QUITTING')
	
	mito_genome, mito_seq = list(fasta.items())[0]
	mito_length = len(mito_seq)
	
	if(write_files):
		make_folder(of + "/fasta/")
		make_folder(of + "/final/")
	
	newfastaf = of + "/fasta/" + mito_genome + ".fasta"
	
	# Need a special logic to potentially over-write the existing 
	# mito fasta file if they wind up not being the same from guess/checking hg19
	writeFA = False
	if os.path.exists(newfastaf):
		if filecmp.cmp(fastaf, newfastaf,shallow=False):
			pass
		else:
			writeFA = True
	else:
		writeFA = True
		
	if writeFA and write_files:
			
		shutil.copyfile(fastaf, newfastaf)
		fastaf = newfastaf
		pysam.faidx(fastaf)
	
		with open(of + "/final/" + mito_genome + "_refAllele.txt", 'w') as f:
			b = 1
			for base in mito_seq:
				f.write(str(b) + "\t" + base + "\n")
				b += 1
			f.close()
	return(fastaf, mito_genome, mito_length)
	

def make_folder(folder):
	"""
	Function to only make a given folder if it does not already exist
	"""
	if not os.path.exists(folder):
		os.makedirs(folder)

def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def split_barcodes_file(barcode_file, nsamples, output):
	"""
	Function to only make a given folder if it does not already exist
	"""
	n_samples_observed = file_len(barcode_file)
	
	# See if we need to do anything... if user didn't specify or we have relatively few samples, just return the barcode file back
	if(n_samples_observed < nsamples or nsamples == 0):
		return([barcode_file])
	else:
		# Need to split files into a maximum of n samples
		total_files = math.ceil(n_samples_observed / nsamples)
		lines_per_file = nsamples
		
		# Set up a temporary output folder to route files to
		full_output_folder = output + "/temp" + "/barcode_files"
		make_folder(full_output_folder)
		smallfile = None
		counter = 0
		with open(barcode_file) as bigfile:
			for lineno, line in enumerate(bigfile):
				if lineno % lines_per_file == 0:
					if smallfile:
						smallfile.close()
					counter = counter + 1
					small_filename = full_output_folder + "/barcodes." + str(counter) + ".txt" 
					smallfile = open(small_filename, "w")
				smallfile.write(line)
		if smallfile:
			smallfile.close()
		barcodes_files = [full_output_folder + "/barcodes." + str(x) + ".txt" for x in list(range(1, total_files + 1))]
		return(barcodes_files)



# https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python	
def available_cpu_count():
    """
    Number of available virtual or physical CPUs on this system, i.e.
    user/real as output by time(1) when called with an optimally scaling
    userspace-only program
	"""
    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
                      open('/proc/self/status').read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # http://code.google.com/p/psutil/
    try:
        import psutil
        return psutil.cpu_count()   # psutil.NUM_CPUS on old versions
    except (ImportError, AttributeError):
        pass

    # POSIX
    try:
        res = int(os.sysconf('SC_NPROCESSORS_ONLN'))

        if res > 0:
            return res
    except (AttributeError, ValueError):
        pass

    # Windows
    try:
        res = int(os.environ['NUMBER_OF_PROCESSORS'])

        if res > 0:
            return res
    except (KeyError, ValueError):
        pass

    # jython
    try:
        from java.lang import Runtime
        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()
        if res > 0:
            return res
    except ImportError:
        pass

    # BSD
    try:
        sysctl = subprocess.Popen(['sysctl', '-n', 'hw.ncpu'],
                                  stdout=subprocess.PIPE)
        scStdout = sysctl.communicate()[0]
        res = int(scStdout)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        res = open('/proc/cpuinfo').read().count('processor\t:')

        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudoDevices = os.listdir('/devices/pseudo/')
        res = 0
        for pd in pseudoDevices:
            if re.match(r'^cpuid@[0-9]+$', pd):
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            dmesg = open('/var/run/dmesg.boot').read()
        except IOError:
            dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
            dmesg = dmesgProcess.communicate()[0]

        res = 0
        while '\ncpu' + str(res) + ':' in dmesg:
            res += 1

        if res > 0:
            return res
    except OSError:
        pass

    raise Exception('Can not determine number of CPUs on this system')
    



#!/usr/bin/python

###################################################
# Summarizes the total number of reads per position / strand
###################################################

import sys
import re
import os
import pysam
import numpy as np
from collections import defaultdict

bam_file = sys.argv[1]
barcodes_file = sys.argv[2]
out_pre = sys.argv[3]
max_bp = int(sys.argv[4])
base_qual = float(sys.argv[5])
fasta_file = sys.argv[6]
alignment_quality = float(sys.argv[7])
barcode_tag = sys.argv[8]

# Import barcodes
with open(barcodes_file) as barcode_file_handle:
    content = barcode_file_handle.readlines()
bcs = [x.strip() for x in content]
dna_letters = ['A','C','G','T']

def getBarcode(intags):
	'''
	Parse out the barcode per-read
	'''
	for tg in intags:
		if(barcode_tag == tg[0]):
			return(tg[1])
	return("NA")

# organize reads into a dict where key is [cellbc]$[readname]
bam_input = [x for x in pysam.AlignmentFile(bam_file, "rb")]
ordered_bam_input = defaultdict(list)
for read in bam_input:
	cell_barcode = getBarcode(read.tags)
	if cell_barcode != "NA":
		read_key = '$'.join([cell_barcode, read.query_name])
		ordered_bam_input[read_key].append(read)

# Dimension cell x position x letter x strand
# Coverage associated with the bases
ca =  np.zeros((len(bcs), max_bp, 4, 2), dtype=int)
for read_key in ordered_bam_input:
	# disregard singlets and multiplets
	if len(ordered_bam_input[read_key]) != 2:
		continue
	
	# identify fwd and rev in a pair
	read0, read1 = ordered_bam_input[read_key]
	if read0.is_reverse and not read1.is_reverse:
		fwd_read, rev_read = read1, read0
	elif not read0.is_reverse and read1.is_reverse:
		fwd_read, rev_read = read0, read1
	else:
		# disregard a pair if both are the same strand
		continue
		
	# gather what we need
	fwd_seq, rev_seq = fwd_read.query_sequence, rev_read.query_sequence
	fwd_quality, rev_quality = np.array(fwd_read.query_qualities), np.array(rev_read.query_qualities)
	fwd_align_qual_read, rev_align_qual_read = fwd_read.mapping_quality, rev_read.mapping_quality

	cell_barcode = read_key.split('$')[0]
	c_idx = bcs.index(cell_barcode)

	# check alignment quality
	if fwd_align_qual_read > alignment_quality and rev_align_qual_read > alignment_quality:
		# partition the pair into fwd-only, overlap, and rev-only
		overlap_length = fwd_read.get_overlap(rev_read.reference_start, rev_read.reference_end)
		if overlap_length == 0:
			fwd_use_idx = np.arange(len(fwd_seq))
			rev_use_idx = np.arange(len(rev_seq))
		else:
			# choose which strand to use in the overlap region based on quality score
			fwd_only_end = len(fwd_seq) - overlap_length
			rev_only_start = overlap_length
			fwd_overlap_quality = fwd_quality[fwd_only_end:]
			rev_overlap_quality = rev_quality[:rev_only_start]
			fwd_overlap_use_idx = np.where(fwd_overlap_quality > rev_overlap_quality)[0]
			rev_overlap_use_idx = np.where(fwd_overlap_quality < rev_overlap_quality)[0]
			
			# evenly assign bases with equal quality in overlap region
			equal_overlap_idx = np.where(fwd_overlap_quality == rev_overlap_quality)[0]
			equal_split = int(np.floor(len(equal_overlap_idx)/2))
			fwd_overlap_use_idx = np.concatenate([fwd_overlap_use_idx, equal_overlap_idx[:equal_split]])
			rev_overlap_use_idx = np.concatenate([rev_overlap_use_idx, equal_overlap_idx[equal_split:]])

			# merge the exclusive region and use idx in overlap region
			fwd_use_idx = np.concatenate([np.arange(fwd_only_end), fwd_overlap_use_idx + fwd_only_end])
			rev_use_idx = np.concatenate([rev_overlap_use_idx, np.arange(rev_only_start, len(rev_seq))])
	
	elif fwd_align_qual_read <= alignment_quality and rev_align_qual_read <= alignment_quality:
		# use none for either
		fwd_use_idx = np.array([])
		rev_use_idx = np.array([])
	
	elif fwd_align_qual_read > alignment_quality and rev_align_qual_read <= alignment_quality:
		# use none of rev and all of fwd
		fwd_use_idx = np.arange(len(fwd_seq))
		rev_use_idx = np.array([])
	
	elif fwd_align_qual_read <= alignment_quality and rev_align_qual_read > alignment_quality:
		# use all of rev and none of fwd
		fwd_use_idx = np.array([])
		rev_use_idx = np.arange(len(rev_seq))

	# handle fwd region
	fwd_aligned_pairs = fwd_read.get_aligned_pairs(True)
	fwd_region = [pair for pair in fwd_aligned_pairs if pair[0] in fwd_use_idx]
	for qpos, refpos in fwd_region:
		if (refpos is not None) and (fwd_quality[qpos] > base_qual) and (fwd_seq[qpos] in dna_letters):
			l_idx = dna_letters.index(fwd_seq[qpos])
			ca[c_idx, refpos, l_idx, 0] += 1

	# handle rev region
	rev_aligned_pairs = rev_read.get_aligned_pairs(True)
	rev_region = [pair for pair in rev_aligned_pairs if pair[0] in rev_use_idx]
	for qpos, refpos in rev_region:
		if (refpos is not None) and (rev_quality[qpos] > base_qual) and (rev_seq[qpos] in dna_letters):
			l_idx = dna_letters.index(rev_seq[qpos])
			ca[c_idx, refpos, l_idx, 1] += 1


# Function to write the slice of the matrix that is associated with the 
def writeSparseMatrixLetter(letter, letter_idx):
	out_file_fn = out_pre + "."+letter+".txt"
	with open(out_file_fn,"w") as file_handle_fn:
		for cell_idx, cell_name in enumerate(bcs):
			
			# Pull out the stranded counts
			fw_vec = ca[cell_idx,:,letter_idx,0].ravel()
			rev_vec = ca[cell_idx,:,letter_idx,1].ravel()
			
			# Write each position
			for i in range(0,int(max_bp)):
				if(fw_vec[i] > 0 or rev_vec[i] > 0):
					file_handle_fn.write(str(i+1)+","+cell_name+","+str(fw_vec[i])+","+str(rev_vec[i])+"\n")

writeSparseMatrixLetter("A", 0)
writeSparseMatrixLetter("C", 1)
writeSparseMatrixLetter("G", 2)
writeSparseMatrixLetter("T", 3)

# Export the per-base coverage for the thrill of it and the depth
out_file_depth = out_pre.replace("/temp/sparse_matrices/", "/qc/depth/") + ".depth.txt"
out_file_coverage= out_pre + ".coverage.txt"
with open(out_file_coverage,"w") as file_handle_cov:
	with open(out_file_depth,"w") as file_handle_depth:
	
		# Loop over cells
		for cell_idx, cell_name in enumerate(bcs):
		
			# Pull out the summed counts per cell per position
			cov_vec = np.sum(ca[cell_idx,:,:,:], axis = (1,2)).tolist()
			depth = round(sum(cov_vec)/len(cov_vec),2)
			# Write each position
			for i in range(0,int(max_bp)):
				if(cov_vec[i] > 0):
					file_handle_cov.write(str(i+1)+","+cell_name+","+str(cov_vec[i])+"\n")
			
			# Now write the depth
			file_handle_depth.write(cell_name+"\t"+str(depth)+"\n")







import click
import os
import os.path
import sys
import shutil
import random
import string
import itertools
import time
import pysam
import math
import glob

from pkg_resources import get_distribution
from subprocess import call, check_call
from .mgatkHelp import *
from ruamel.yaml import YAML
from itertools import repeat
from ruamel.yaml.scalarstring import SingleQuotedScalarString as sqs
from multiprocessing import Pool

@click.command()
@click.version_option()
@click.argument('mode', type=click.Choice(['bcall', 'call', 'tenx', 'check', 'support', 'remove-background']))
@click.option('--input', '-i', default = ".", required=True, help='Input; either directory of singular .bam file; see documentation. REQUIRED.')
@click.option('--output', '-o', default="mgatk_out", help='Output directory for analysis required for `call` and `bcall`. Default = mgatk_out')
@click.option('--name', '-n', default="mgatk",  help='Prefix for project name. Default = mgatk')

@click.option('--mito-genome', '-g', default = "rCRS", required=True, help='mitochondrial genome configuration. Choose hg19, hg38, mm10, (etc.) or a custom .fasta file; see documentation. Default = rCRS.')
@click.option('--ncores', '-c', default = "detect", help='Number of cores to run the main job in parallel.')

@click.option('--cluster', default = "",  help='Message to send to Snakemake to execute jobs on cluster interface; see documentation.')
@click.option('--jobs', default = "0",  help='Max number of jobs to be running concurrently on the cluster interface.')

@click.option('--barcode-tag', '-bt', default = "X",  help='Read tag (generally two letters) to separate single cells; valid and required only in `bcall` mode.')
@click.option('--barcodes', '-b', default = "",  help='File path to barcodes that will be extracted; useful only in `bcall` mode. If none supplied, mgatk will learn abundant barcodes from the bam file (threshold defined by the -mb tag).')
@click.option('--min-barcode-reads', '-mb', default = 1000,  help='Minimum number of mitochondrial reads for a barcode to be genotyped; useful only in `bcall` mode; will not overwrite the `--barcodes` logic. Default = 1000.')

@click.option('--NHmax', default = 1, help='Maximum number of read alignments allowed as governed by the NH flag. Default = 1.')
@click.option('--NMmax', default = 4, help='Maximum number of paired mismatches allowed represented by the NM/nM tags. Default = 4.')

@click.option('--keep-duplicates', '-kd', is_flag=True, help='Retained dupliate (presumably PCR) reads')
@click.option('--umi-barcode', '-ub', default = "",  help='Read tag (generally two letters) to specify the UMI tag when removing duplicates for genotyping.')

@click.option('--handle-overlap', '-ho', is_flag=True, help='Only count each base in the overlap region between a pair of reads once')
@click.option('--low-coverage-threshold', '-lc', default=10, help='Variant count for each cell will be ignored below this when calculating VMR')

@click.option('--max-javamem', '-jm', default = "8000m", help='Maximum memory for java for running duplicate removal per core. Default = 8000m.')

@click.option('--proper-pairs', '-pp', is_flag=True, help='Require reads to be properly paired.')

@click.option('--base-qual', '-q', default = 0, help='Minimum base quality for inclusion in the genotype count. Default = 0.')
@click.option('--alignment-quality', '-aq', default = 0, help='Minimum alignment quality to include read in genotype. Default = 0.')
@click.option('--emit-base-qualities', '-eb', is_flag=True, help='Output mean base quality per alt allele as part of the final output.')

@click.option('--nsamples', '-ns', default = 7000, help='The number of samples / cells to be processed per iteration; Default = 7000. Supply 0 to try all.')

@click.option('--keep-samples', '-k', default="ALL", help='Comma separated list of sample names to keep; ALL (special string) by default. Sample refers to basename of .bam file')
@click.option('--ignore-samples', '-x', default="NONE", help='Comma separated list of sample names to ignore; NONE (special string) by default. Sample refers to basename of .bam file')

@click.option('--keep-temp-files', '-z', is_flag=True, help='Add this flag to keep all intermediate files.')
@click.option('--keep-qc-bams', '-qc', is_flag=True, help='Add this flag to keep the quality-controlled bams after processing.')

@click.option('--skip-R', '-sr', is_flag=True, help='Generate plain-text only output. Otherwise, this generates a .rds obejct that can be immediately read into R for downstream analysis.')
@click.option('--snake-stdout', '-so', is_flag=True, help='Write snakemake log to sdout rather than a file. May be necessary for certain HPC environments.')

@click.option('--ncells_fg', '-nfg', default = 1000,  help='number of "foreground" cells to use for CellBender. Default = 1000.')
@click.option('--ncells_bg', '-nbg', default = 20000,  help='number of "background" cells to use for CellBender. Default = 20000.')

def main(mode, input, output, name, mito_genome, ncores,
	cluster, jobs, barcode_tag, barcodes, min_barcode_reads,
	nhmax, nmmax, keep_duplicates, umi_barcode, handle_overlap, low_coverage_threshold,
	max_javamem, proper_pairs, base_qual, alignment_quality, emit_base_qualities,
	nsamples, keep_samples, ignore_samples,
	     keep_temp_files, keep_qc_bams, skip_r, snake_stdout, ncells_fg, ncells_bg):
	
	"""
	mgatk: a mitochondrial genome analysis toolkit. \n
	MODE = ['bcall', 'call', 'tenx', 'check', 'support', 'remove-background'] \n
	See https://github.com/caleblareau/mgatk/wiki for more details.
	"""
	
	script_dir = os.path.dirname(os.path.realpath(__file__))
	cwd = os.getcwd()
	__version__ = get_distribution('mgatk').version
	click.echo(gettime() + "mgatk v%s" % __version__)
	
	# Determine cores
	if(ncores == "detect"):
		ncores = str(available_cpu_count())
	else:
		ncores = str(ncores)
	
	# Now removing duplicates is the default, so recode variable names
	if(keep_duplicates):
		remove_duplicates = False
	else:
		remove_duplicates = True
	
	# Verify dependencies	
	if remove_duplicates:
		check_software_exists("java")
	
	if (mode == "call" or mode == "tenx" or mode == "bcall") and not skip_r:
		check_software_exists("R")
		check_R_packages(["data.table", "SummarizedExperiment", "GenomicRanges", "Matrix"])
	elif (mode=='remove-background'):
		check_software_exists("R")
		check_R_packages(["data.table", "SummarizedExperiment", "GenomicRanges", "Matrix", "optparse", "dplyr", "tidyr", "hdf5r"])
	
	# Determine which genomes are available
	rawsg = glob.glob(script_dir + "/bin/anno/fasta/*.fasta")
	supported_genomes = [x.replace(script_dir + "/bin/anno/fasta/", "").replace(".fasta", "") for x in rawsg]  
	
	if(mode == "support"):
		click.echo(gettime() + "List of built-in genomes supported in mgatk:")
		click.echo(gettime() + str(supported_genomes))
		sys.exit(gettime() + 'Specify one of these genomes or provide your own .fasta file with the --mito-genome flag')
		
	if(mode == "check"):
		click.echo(gettime() + "checking dependencies...")
	
	# Remember that I started off as bcall as this will become overwritten
	wasbcall = False
	continue_check = True
	if(mode == "bcall" or mode == "tenx" or mode == "check"):
	
		if(barcode_tag == "X"):
			sys.exit('ERROR: in `'+mode+'` mode, must specify a valid read tag ID (generally two letters).')
			
		# Input argument is assumed to be a .bam file
		filename, file_extension = os.path.splitext(input)
		if(file_extension != ".bam"):
			sys.exit('ERROR: in `'+mode+'` mode, the input should be an individual .bam file.')
		if not os.path.exists(input):
			sys.exit('ERROR: No file found called "' + input + '"; please specify a valid .bam file.')
		if not os.path.exists(input + ".bai"):
			click.echo(gettime() + "Attempting to index: " + input)
			pysam.index(input)
			
			if not os.path.exists(input + ".bai"):
				sys.exit('ERROR: index your input .bam file for `bcall` mode.')
		click.echo(gettime() + "Found bam file: " + input + " for genotyping.")
		
		# Determine whether or not we have been supplied barcodes
		barcode_known = False
		if (os.path.exists(barcodes)) and (barcodes != ""):
			click.echo(gettime() + "Found file of barcodes to be parsed: " + barcodes)
			barcode_known = True
		else:
			if(mode == "tenx"):
				sys.exit(gettime() + 'Must specify a known barcode list with `tenx` mode')
			click.echo(gettime() + "Will determine barcodes with at least: " + str(min_barcode_reads) + " mitochondrial reads.")
			
		# Make temporary directory of inputs
		of = output; tf = of + "/temp"; bcbd = tf + "/barcoded_bams" # bcdb = barcoded bam directory
		folders = [of, tf, bcbd, of + "/final"]
		mkfolderout = [make_folder(x) for x in folders]
		
		# Handle fasta requirements
		fastaf, mito_chr, mito_length = handle_fasta_inference(mito_genome, supported_genomes, script_dir, mode, of)
		idxs = pysam.idxstats(input).split("\n")
		
		# Handle common mtDNA reference genome errors
		bam_length = 0
		for i in idxs:
			if(i.split("\t")[0] == mito_chr):
				bam_length = int(i.split("\t")[1])
		
		if(mito_length == bam_length):
			click.echo(gettime() + "User specified mitochondrial genome matches .bam file")
		elif(bam_length == 16569):
			click.echo(gettime() + "User specified mitochondrial genome does NOT match .bam file; using rCRS instead (length == 16569)")
			fastaf, mito_chr, mito_length = handle_fasta_inference("rCRS", supported_genomes, script_dir, mode, of)
		elif(bam_length == 16571):
			click.echo(gettime() + "User specified mitochondrial genome does NOT match .bam file; using hg19 instead (length == 16571)")
			fastaf, mito_chr, mito_length = handle_fasta_inference("hg19", supported_genomes, script_dir, mode, of)
		else:
			click.echo(gettime() + "User specified mitochondrial genome does NOT match .bam file; correctly specify reference genome or .fasta file")
			quit()
		
		# Actually call the external script based on user input
		if(not barcode_known):
			barc_quant_file = of + "/final/barcodeQuants.tsv"
			passing_barcode_file = of + "/final/passingBarcodes.tsv"
			find_barcodes_py = script_dir + "/bin/python/find_barcodes.py"
			
			pycall = " ".join(['python', find_barcodes_py, input, bcbd, barcode_tag, str(min_barcode_reads), mito_chr, barc_quant_file, passing_barcode_file])
			os.system(pycall)
			barcodes = passing_barcode_file

		# Potentially split the valid barcodes into smaller files if we need to
		if(mode == "bcall"):
			barcode_files = split_barcodes_file(barcodes, nsamples, output)
			split_barcoded_bam_py = script_dir + "/bin/python/split_barcoded_bam.py"
			
			# Loop over the split sample files
			for i in range(len(barcode_files)):
				one_barcode_file = barcode_files[i]
				pycall = " ".join(['python', split_barcoded_bam_py, input, bcbd, barcode_tag, one_barcode_file, mito_chr])
				os.system(pycall)
				
			# Update everything to appear like we've just set `call` on the set of bams
			mode = "call"
			input = bcbd 
			wasbcall = True
		
		#
		if(mode == "check" or mode == "tenx"):
			barcode_files = split_barcodes_file(barcodes, math.ceil(file_len(barcodes)/int(ncores)), output)
			samples = [os.path.basename(os.path.splitext(sample)[0]) for sample in barcode_files] 
			samplebams = [of + "/temp/barcoded_bams/" + sample + ".bam" for sample in samples]
			
			if(umi_barcode == ""):
				umi_barcode = "XX"
			
			# Enact the split in a parallel manner if necessary
			if(mode == "tenx"):
				pool = Pool(processes=int(ncores))
				pmblah = pool.starmap(split_chunk_file, zip(barcode_files, repeat(script_dir), repeat(input), repeat(bcbd), repeat(barcode_tag), repeat(mito_chr), repeat(umi_barcode)))
				pool.close()
				
			umi_barcode = "MU"
			continue_check = False
		
	
		click.echo(gettime() + "Finished determining/splitting barcodes for genotyping.")
		
	# -------------------------------
	# Determine samples for analysis
	# -------------------------------
	if((mode == "check" or mode == "call") and continue_check):
	
		bams = []
		bams = glob.glob(input + '/*.bam')

		if bams[0] == '':
			sys.exit('ERROR: Could not import any samples from the user specification; check flags, logs and input configuration; QUITTING')
	
		samples = []
		samplebams = []
		
		if(not wasbcall):
			fastaf, mito_chr, mito_length = handle_fasta_inference(mito_genome, supported_genomes, script_dir, mode, output, write_files = False)
		
		# Loop over bam files
		for bam in bams:
			base=os.path.basename(bam)
			basename=os.path.splitext(base)[0]
			samples.append(basename)
			samplebams.append(bam)
			
		# parallel process to ensure we have .bai files for each bam
		pool = Pool(processes=int(ncores))
		pm = pool.map(verify_bai, samplebams)
		pool.close()
		
		samples_fail = []
		for i in range(len(samples)):
			sample = samples[i]
			bam = samplebams[i]
			if( not verify_sample_mitobam(bam, mito_chr, mito_length)):
				samples_fail.append(sample)
				
		if(keep_samples != "ALL"):
			keeplist = keep_samples.split(",")
			click.echo(gettime() + "Intersecting detected samples with user-retained ones: " + keep_samples)
			keepidx = findIdx(samples, keeplist)
			samples = [samples[i] for i in keepidx]
			samplebams = [samplebams[i] for i in keepidx]
		
		if(ignore_samples != "NONE"):
			iglist = ignore_samples.split(",")
			click.echo(gettime() + "Will remove samples from processing:" + ignore_samples)
			rmidx = findIdx(samples, iglist)
			for index in sorted(rmidx, reverse=True):
				del samples[index]
				del samplebams[index]
				
		if(len(samples_fail) > 0):
			click.echo(gettime() + "NOTE: the samples below either have 0 mtDNA reads at the specified chromosome or are mapped to an incorrectly specified reference mitochondrial genome")
			click.echo(gettime() + "Will remove samples from processing:" )
			rmidx = findIdx(samples, samples_fail)
			for index in sorted(rmidx, reverse=True):
				print("REMOVED: ", samples[index])
				del samples[index]
				del samplebams[index]
			
		if not len(samples) > 0:
			sys.exit('ERROR: Could not import any samples from the user specification. \nERROR: check flags, logs, and input configuration (including reference mitochondrial genome); \nQUITTING')
	
		nsamplesNote = "mgatk will process " + str(len(samples)) + " samples"
		
	if(mode == "check"):
		# Exit gracefully
		sys.exit(gettime() + "mgatk check passed!")
			

	if(mode == "call" or mode == "tenx"):
	
		# Make all of the output folders if necessary
		of = output; tf = of + "/temp"; qc = of + "/qc"; logs = of + "/logs"
		folders = [logs, of + "/logs/filterlogs", of + "/fasta", of + "/.internal",
			 of + "/.internal/parseltongue", of + "/.internal/samples", of + "/final", 
			 tf, tf + "/ready_bam", tf + "/temp_bam", tf + "/sparse_matrices", tf + "/quality",
			 qc, qc + "/quality", qc + "/depth"]

		mkfolderout = [make_folder(x) for x in folders]
		
		#-------------------
		# Handle .fasta file
		#-------------------
		if((mode == "call" and wasbcall == False)):
			fastaf, mito_genmito_chrome, mito_length = handle_fasta_inference(mito_genome, supported_genomes, script_dir, mode, of)
			print(gettime() + "Found designated mitochondrial chromosome: %s" % mito_chr)
			
		if(mode == "call" or mode == "tenx"):
			# Logging		
			logf = open(output + "/logs" + "/base.mgatk.log", 'a')
			click.echo(gettime() + "Starting analysis with mgatk", logf)
			
			if(mode == "call"):
				click.echo(gettime() + nsamplesNote, logf)

		if (remove_duplicates):
				make_folder(of + "/logs/rmdupslogs")
	
		# Create internal README files 
		if not os.path.exists(of + "/.internal/README"):
			with open(of + "/.internal/README" , 'w') as outfile:
				outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/parseltongue/README"):	
			with open(of + "/.internal/parseltongue/README" , 'w') as outfile:
				outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")
		if not os.path.exists(of + "/.internal/samples/README"):
			with open(of + "/.internal" + "/samples" + "/README" , 'w') as outfile:
				outfile.write("This folder creates samples to be interpreted by Snakemake; don't modify it.\n\n")
	
		# Set up sample bam plain text file
		for i in range(len(samples)):
			with open(of + "/.internal/samples/" + samples[i] + ".bam.txt" , 'w') as outfile:
				outfile.write(samplebams[i])
		
		click.echo(gettime() + "Genotyping samples with "+ncores+" threads")
		
		# add sqs to get .yaml to play friendly https://stackoverflow.com/questions/39262556/preserve-quotes-and-also-add-data-with-quotes-in-ruamel
		dict1 = {'input_directory' : sqs(input), 'output_directory' : sqs(output), 'script_dir' : sqs(script_dir),
			'fasta_file' : sqs(fastaf), 'mito_chr' : sqs(mito_chr), 'mito_length' : sqs(mito_length), 'name' : sqs(name),
			'base_qual' : sqs(base_qual), 'remove_duplicates' : sqs(remove_duplicates), 'handle_overlap' : sqs(handle_overlap),
			'low_coverage_threshold' : sqs(low_coverage_threshold), 'barcode_tag' : sqs(barcode_tag), 'umi_barcode' : sqs(umi_barcode),
			'alignment_quality' : sqs(alignment_quality), 'emit_base_qualities' : sqs(emit_base_qualities),
			'proper_paired' : sqs(proper_pairs), 'NHmax' : sqs(nhmax), 'NMmax' : sqs(nmmax), 'max_javamem' : sqs(max_javamem)}
		
		# Potentially submit jobs to cluster
		snakeclust = ""
		njobs = int(jobs)
		if(njobs > 0 and cluster != ""):
			snakeclust = " --jobs " + jobs + " --cluster '" + cluster + "' "
			click.echo(gettime() + "Recognized flags to process jobs on a computing cluster.", logf)
			
		click.echo(gettime() + "Processing samples with "+ncores+" threads", logf)
		
		y_s = of + "/.internal/parseltongue/snake.scatter.yaml"
		with open(y_s, 'w') as yaml_file:
			yaml=YAML()
			yaml.default_flow_style = False
			yaml.dump(dict1, yaml_file)
		
		cp_call = "cp " + y_s +  " " + logs + "/" + name + ".parameters.txt"
		os.system(cp_call)
	
		if(mode == "call"):
			
			# Execute snakemake
			snake_log = logs + "/" + name + ".snakemake_scatter.log"
			
			snake_log_out = ""
			if not snake_stdout:
				snake_log_out = ' &>' + snake_log
				
			snakecmd_scatter = 'snakemake'+snakeclust+' --snakefile ' + script_dir + '/bin/snake/Snakefile.Scatter --cores '+ncores+' --config cfp="'  + y_s + '"' + snake_log_out
			os.system(snakecmd_scatter)
			
		elif(mode == "tenx"):
			
			# Execute snakemake
			snake_log = logs + "/" + name + ".snakemake_tenx.log"
			
			snake_log_out = ""
			if not snake_stdout:
				snake_log_out = ' &>' + snake_log
				
			snakecmd_tenx = 'snakemake'+snakeclust+' --snakefile ' + script_dir + '/bin/snake/Snakefile.tenx --cores '+ncores+' --config cfp="'  + y_s + '"'+ snake_log_out
			os.system(snakecmd_tenx)
		
		click.echo(gettime() + "mgatk successfully processed the supplied .bam files", logf)

	
	#-------
	# Gather
	#-------
	if(mode == "call"):
		
		mgatk_directory = output
		dict2 = {'mgatk_directory' : sqs(mgatk_directory), 'name' : sqs(name),
			'script_dir' : sqs(script_dir)}
		y_g = mgatk_directory + "/.internal/parseltongue/snake.gather.yaml"
		with open(y_g, 'w') as yaml_file:
			yaml=YAML()
			yaml.default_flow_style = False
			yaml.dump(dict2, yaml_file)
		
		# Snakemake gather
		snake_log = logs + "/" + name + ".snakemake_gather.log"
		
		snake_log_out = ""
		if not snake_stdout:
			snake_log_out = ' &>' + snake_log 
			
		snakecmd_gather = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.Gather --cores 1 --config cfp="' + y_g + '"' + snake_log_out
		os.system(snakecmd_gather)
	
	if(mode == "call" or mode == "tenx"):

		# Make .rds file from the output
		Rcall = "Rscript " + script_dir + "/bin/R/toRDS.R " + output + "/final " + name
		os.system(Rcall)
		click.echo(gettime() + "Successfully created final output files", logf)

	if(mode == "remove-background"):
			
		of = output
		check_software_exists("cellbender")
		check_software_exists("Rscript")
		
		logf = open(output + "/logs" + "/bg.mgatk.log", 'a')
		# prepare mgatk output for CellBender
		# check that software exists
		
		cellbender_input_dir = output + "/cellbender_input"
		make_folder(cellbender_input_dir)
		# call prepare_cellbender.R to convert mgatk output to cellbender input for variable positions (specify cutoffs on --n_cells_conf_detected, --strand_correlation and --vmr as well?)"
		Rcall = "Rscript " + script_dir + "/bin/R/prepare_cellbender.R -i " + output + "/final -o " + cellbender_input_dir
		os.system(Rcall)
		#print(Rcall)
		click.echo(gettime() + "Prepared input for CellBender", logf)

		# run CellBender
		# check that software exists
		
		cellbender_output_dir = output + "/cellbender_output"
		make_folder(cellbender_output_dir)
		
        # pass other arguments to cellbender?
		cellbender_cmd = "cellbender remove-background --input " + output + "/cellbender_input --output " + output + "/cellbender_output/" + name + ".h5 --expected-cells " + str(ncells_fg) + " --total-droplets-included " + str(ncells_bg) + " --fpr 0.01 --epochs 100 --low-count-threshold 1"
		os.system(cellbender_cmd)
		#print(cellbender_cmd)
		click.echo(gettime() + "Finished CellBender run", logf)

		# convert CellBender output (here we could restrict the output to good cells only with the --cells argument, expects the barcodes.tsv file)
		Rcall = "Rscript " + script_dir + "/bin/R/convert_cellbender_output.R -i " + output + "/final -c " + output + "/cellbender_output/" + name + ".h5 -o " + output + "/final/" + name + "_filtered.rds"
		os.system(Rcall)
		#print(Rcall)
		click.echo(gettime() + "Converted CellBender output", logf)

		if keep_temp_files:
			click.echo(gettime() + "Temporary files not deleted since --keep-temp-files was specified.", logf)
		else:
			shutil.rmtree(of + "/cellbender_input")
			shutil.rmtree(of + "/cellbender_output")
			click.echo(gettime() + "Intermediate files successfully removed.", logf)
	
		# Suspend logging
		logf.close()

	#--------
	# Cleanup
	#--------
	if(mode == "call" or mode == "tenx"):
		if keep_qc_bams:
			click.echo(gettime() + "Final bams retained since --keep-qc-bams was specified.", logf)
			dest = shutil.move(of + "/temp/ready_bam", of + "/qc_bam")	

		if keep_temp_files:
			click.echo(gettime() + "Temporary files not deleted since --keep-temp-files was specified.", logf)
		else:
			shutil.rmtree(of+ "/fasta")
			shutil.rmtree(of + "/.internal")
			shutil.rmtree(of + "/temp")
			click.echo(gettime() + "Intermediate files successfully removed.", logf)
		
		# Suspend logging
		logf.close()

	
	



#!/usr/bin/python

###################################################
# Summarizes the total number of reads per position
###################################################

import sys
import re
import os
import pysam

bamfile = sys.argv[1]
outpre = sys.argv[2]
mito_genome = sys.argv[3]
maxBP = sys.argv[4]
base_qual = float(sys.argv[5])
sample = sys.argv[6]
fasta_file = sys.argv[7]
alignment_quality = float(sys.argv[8])
emit_base_qualities = sys.argv[9]

# Export Functions
def writeSparseMatrix(mid, vec):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec[i])+"\n")


def writeSparseMatrix2(mid, vec1, vec2):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec1[i] > 0 or vec2[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec1[i])+","+str(vec2[i])+"\n")

def writeSparseMatrix4(mid, vec1, vec2, vec3, vec4):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec1[i] > 0 or vec3[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec1[i])+","+str(vec2[i])+","+str(vec3[i])+","+str(vec4[i])+"\n")


n = int(maxBP)

# initialize with a pseudo count to avoid dividing by zero
countsA_fw = [0.00000001] * n 
countsC_fw = [0.00000001] * n 
countsG_fw = [0.00000001] * n 
countsT_fw = [0.00000001] * n 

qualA_fw = [0.0] * n
qualC_fw = [0.0] * n
qualG_fw = [0.0] * n
qualT_fw = [0.0] * n

countsA_rev = [0.00000001] * n 
countsC_rev = [0.00000001] * n 
countsG_rev = [0.00000001] * n 
countsT_rev = [0.00000001] * n 

qualA_rev = [0.0] * n
qualC_rev = [0.0] * n
qualG_rev = [0.0] * n
qualT_rev = [0.0] * n

bam2 = pysam.AlignmentFile(bamfile, "rb")
for read in bam2:
	seq = read.seq
	reverse = read.is_reverse
	quality = read.query_qualities
	align_qual_read = read.mapping_quality
	for qpos, refpos in read.get_aligned_pairs(True):
		if qpos is not None and refpos is not None and align_qual_read > alignment_quality:
			if(seq[qpos] == "A" and quality[qpos] > base_qual):
				if(not reverse):
					qualA_fw[refpos] += quality[qpos]
					countsA_fw[refpos] += 1
				else:
					qualA_rev[refpos] += quality[qpos]
					countsA_rev[refpos] += 1
			elif(seq[qpos] == "C" and quality[qpos] > base_qual):
				if(not reverse):
					qualC_fw[refpos] += quality[qpos]
					countsC_fw[refpos] += 1
				else:
					qualC_rev[refpos] += quality[qpos]
					countsC_rev[refpos] += 1
			elif(seq[qpos] == "G" and quality[qpos] > base_qual):
				if(not reverse):
					qualG_fw[refpos] += quality[qpos]
					countsG_fw[refpos] += 1
				else:
					qualG_rev[refpos] += quality[qpos]
					countsG_rev[refpos] += 1
			elif(seq[qpos] == "T" and quality[qpos] > base_qual):
				if(not reverse):
					qualT_fw[refpos] += quality[qpos]
					countsT_fw[refpos] += 1
				else:
					qualT_rev[refpos] += quality[qpos]
					countsT_rev[refpos] += 1
			
meanQualA_fw = [round(x/y,1) for x, y in zip(qualA_fw, countsA_fw)]
meanQualC_fw = [round(x/y,1) for x, y in zip(qualC_fw, countsC_fw)]
meanQualG_fw = [round(x/y,1) for x, y in zip(qualG_fw, countsG_fw)]
meanQualT_fw = [round(x/y,1) for x, y in zip(qualT_fw, countsT_fw)]

countsA_fw = [ int(round(elem)) for elem in countsA_fw ]
countsC_fw = [ int(round(elem)) for elem in countsC_fw ]
countsG_fw = [ int(round(elem)) for elem in countsG_fw ]
countsT_fw = [ int(round(elem)) for elem in countsT_fw ]

meanQualA_rev = [round(x/y,1) for x, y in zip(qualA_rev, countsA_rev)]
meanQualC_rev = [round(x/y,1) for x, y in zip(qualC_rev, countsC_rev)]
meanQualG_rev = [round(x/y,1) for x, y in zip(qualG_rev, countsG_rev)]
meanQualT_rev = [round(x/y,1) for x, y in zip(qualT_rev, countsT_rev)]

countsA_rev = [ int(round(elem)) for elem in countsA_rev ]
countsC_rev = [ int(round(elem)) for elem in countsC_rev ]
countsG_rev = [ int(round(elem)) for elem in countsG_rev ]
countsT_rev = [ int(round(elem)) for elem in countsT_rev ]

# Allele Counts
bam = pysam.AlignmentFile(bamfile, "rb")

if(emit_base_qualities == "True"):
	writeSparseMatrix4("A", countsA_fw, meanQualA_fw, countsA_rev, meanQualA_rev)
	writeSparseMatrix4("C", countsC_fw, meanQualC_fw, countsC_rev, meanQualC_rev)
	writeSparseMatrix4("G", countsG_fw, meanQualG_fw, countsG_rev, meanQualG_rev)
	writeSparseMatrix4("T", countsT_fw, meanQualT_fw, countsT_rev, meanQualT_rev)
else:
	writeSparseMatrix2("A", countsA_fw, countsA_rev)
	writeSparseMatrix2("C", countsC_fw, countsC_rev)
	writeSparseMatrix2("G", countsG_fw, countsG_rev)
	writeSparseMatrix2("T", countsT_fw, countsT_rev)

zipped_list = zip(list(countsA_fw),list(countsC_fw),list(countsG_fw),list(countsT_fw), list(countsA_rev),list(countsC_rev),list(countsG_rev),list(countsT_rev))
sums = [sum(item) for item in zipped_list]
writeSparseMatrix("coverage", sums)



import re
import os
import pysam
from collections import Counter
import sys
from optparse import OptionParser
# Parse out data
opts = OptionParser()
usage = "usage: %prog [options] [inputs] Software to process 1 bam file with specific parameters defined"
opts = OptionParser(usage=usage)
opts.add_option("-i", "--input", help="input single cell bam file")
opts.add_option("-o", "--output", help="output filepath")

# Parse options from command line
options, arguments = opts.parse_args()
inbam = options.input
output_file = options.output
bam_in = pysam.AlignmentFile(inbam, "rb")

#Functions for deletion calling
def process_cigar_for_clip_position(cigar, tuple):
    pos = 0
    # Case 1/2: start of read
    if(cigar[1] == "H" or cigar[2] == "H" or
       cigar[1] == "S" or cigar[2] == "S"):
        pos = tuple[0][1] + 1 #offset 1-base and differentiate 0 pos/noclip
    # Case 2: end of read
    if(cigar[-1] == "H" or cigar[-1] == "S"):
        pos = tuple[-1][1] + 1 #offset 1-base and differentiate 0 pos/noclip
    return(pos)
def left_clip(cigar):
    if(cigar[1] == "H" or cigar[2] == "H" or
       cigar[1] == "S" or cigar[2] == "S"):
        return(1)
    else:
        return(0)
def right_clip(cigar):
    if(cigar[-1] == "S" or cigar[-1] == "H"):
        return(1)
    else:
        return(0)
        
# fuzzy match lev distance
def lev(a, b):
    if not a: return len(b)
    if not b: return len(a)
    return min(lev(a[1:], b[1:])+(a[0] != b[0]), lev(a[1:], b)+1, lev(a, b[1:])+1)
def get_clipped_string(cigar, tuple, seq):
    barcode_5mer="CTGTC"
    barcode_5mer_rc="GACAG"
    # Case 1/2: start of read
    if(cigar[1] == "H" or cigar[2] == "H" or
       cigar[1] == "S" or cigar[2] == "S"):
        pos = tuple[0][0]
        seq_out = seq[:pos][-4:]
        return(lev(barcode_5mer, seq_out) <=1 or
               lev(barcode_5mer_rc, seq_out) <=1 )
    # Case 2: end of read
    if(cigar[-1] == "H" or cigar[-1] == "S"):
        pos = tuple[-1][0]
        seq_out = seq[pos + 1:][:4]
        return(lev(barcode_5mer, seq_out) <=1 or
               lev(barcode_5mer_rc, seq_out) <=1 )
    return(False)
def get_n_clipped(cigar):
    out = "0"
    # Case 1/2: start of read
    if(cigar[1] == "H" or cigar[2] == "H" or
       cigar[1] == "S" or cigar[2] == "S"):
         out = re.split('[A-Z]', cigar)[0]
    # Case 2: end of read
    if(cigar[-1] == "H" or cigar[-1] == "S"):
        out = re.split('[A-Z]', cigar)[-2]
    return(out)

# Open all the output files and spit out the filtered data
# Based on where the matching value originates
clip_pos_count = Counter()
outfile_handle = open(output_file, 'w')
for read in bam_in:
    seq = read.seq
    reverse = read.is_reverse
    cigar_string = read.cigarstring
    positions = read.get_reference_positions()
    tuple = read.get_aligned_pairs(True)
    if(positions and cigar_string and seq):
        start = str(positions[0] + 1) #offset to 1 base
        end = str(positions[-1] + 1) #offset to 1 base
        rc = str(right_clip(cigar_string))
        lc = str(left_clip(cigar_string))
        clip_pos = str(process_cigar_for_clip_position(cigar_string, tuple))
        is_barcode = str(get_clipped_string(cigar_string, tuple, seq))
        clip_pos_count[clip_pos] += 1
        read_name = str(read.query_name)
        n_clipped = get_n_clipped(cigar_string)
        list_of_outs = [start, end, lc, rc, clip_pos, read_name, is_barcode, n_clipped]
        outfile_handle.write("\t".join(list_of_outs) + "\n")
bam_in.close()


#!/usr/bin/python

###################################################
# Summarizes the total number of reads per position
###################################################

import sys
import re
import os
import pysam
import numpy as np
from collections import defaultdict

bamfile = sys.argv[1]
outpre = sys.argv[2]
mito_genome = sys.argv[3]
maxBP = sys.argv[4]
base_qual = float(sys.argv[5])
sample = sys.argv[6]
fasta_file = sys.argv[7]
alignment_quality = float(sys.argv[8])
emit_base_qualities = sys.argv[9]

# Export Functions
def writeSparseMatrix(mid, vec):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec[i])+"\n")


def writeSparseMatrix2(mid, vec1, vec2):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec1[i] > 0 or vec2[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec1[i])+","+str(vec2[i])+"\n")

def writeSparseMatrix4(mid, vec1, vec2, vec3, vec4):
	with open(outpre + "."+mid+".txt","w") as V:
		for i in range(0,int(maxBP)):
			if(vec1[i] > 0 or vec3[i] > 0):
				V.write(str(i+1)+","+sample+","+str(vec1[i])+","+str(vec2[i])+","+str(vec3[i])+","+str(vec4[i])+"\n")


n = int(maxBP)

# initialize with a pseudo count to avoid dividing by zero
countsA_fw = [0.00000001] * n 
countsC_fw = [0.00000001] * n 
countsG_fw = [0.00000001] * n 
countsT_fw = [0.00000001] * n 

qualA_fw = [0.0] * n
qualC_fw = [0.0] * n
qualG_fw = [0.0] * n
qualT_fw = [0.0] * n

countsA_rev = [0.00000001] * n 
countsC_rev = [0.00000001] * n 
countsG_rev = [0.00000001] * n 
countsT_rev = [0.00000001] * n 

qualA_rev = [0.0] * n
qualC_rev = [0.0] * n
qualG_rev = [0.0] * n
qualT_rev = [0.0] * n

# organize reads into a dict where key is readname
bam2 = [x for x in pysam.AlignmentFile(bamfile, "rb")]
ordered_bam2 = defaultdict(list)
for read in bam2:
	ordered_bam2[read.query_name].append(read)

for read_name in ordered_bam2:
	# disregard singlets and multiplets
	if len(ordered_bam2[read_name]) != 2:
		continue
	
	# identify fwd and rev in a pair
	read0, read1 = ordered_bam2[read_name]
	if read0.is_reverse and not read1.is_reverse:
		fwd_read, rev_read = read1, read0
	elif not read0.is_reverse and read1.is_reverse:
		fwd_read, rev_read = read0, read1
	else:
		# disregard a pair if both are the same strand
		continue
	
	# gather what we need
	fwd_seq, rev_seq = fwd_read.query_sequence, rev_read.query_sequence
	fwd_quality, rev_quality = np.array(fwd_read.query_qualities), np.array(rev_read.query_qualities)
	fwd_align_qual_read, rev_align_qual_read = fwd_read.mapping_quality, rev_read.mapping_quality
	
	# check alignment quality
	if fwd_align_qual_read > alignment_quality and rev_align_qual_read > alignment_quality:
		# partition the pair into fwd-only, overlap, and rev-only
		overlap_length = fwd_read.get_overlap(rev_read.reference_start, rev_read.reference_end)
		if overlap_length == 0:
			fwd_use_idx = np.arange(len(fwd_seq))
			rev_use_idx = np.arange(len(rev_seq))
		else:
			# choose which strand to use in the overlap region based on quality score
			fwd_only_end = len(fwd_seq) - overlap_length
			rev_only_start = overlap_length
			fwd_overlap_quality = fwd_quality[fwd_only_end:]
			rev_overlap_quality = rev_quality[:rev_only_start]
			fwd_overlap_use_idx = np.where(fwd_overlap_quality > rev_overlap_quality)[0]
			rev_overlap_use_idx = np.where(fwd_overlap_quality < rev_overlap_quality)[0]
			
			# evenly assign bases with equal quality in overlap region
			equal_overlap_idx = np.where(fwd_overlap_quality == rev_overlap_quality)[0]
			equal_split = int(np.floor(len(equal_overlap_idx)/2))
			fwd_overlap_use_idx = np.concatenate([fwd_overlap_use_idx, equal_overlap_idx[:equal_split]])
			rev_overlap_use_idx = np.concatenate([rev_overlap_use_idx, equal_overlap_idx[equal_split:]])

			# merge the exclusive region and use idx in overlap region
			fwd_use_idx = np.concatenate([np.arange(fwd_only_end), fwd_overlap_use_idx + fwd_only_end])
			rev_use_idx = np.concatenate([rev_overlap_use_idx, np.arange(rev_only_start, len(rev_seq))])
	
	elif fwd_align_qual_read <= alignment_quality and rev_align_qual_read <= alignment_quality:
		# use none for either
		fwd_use_idx = np.array([])
		rev_use_idx = np.array([])
	
	elif fwd_align_qual_read > alignment_quality and rev_align_qual_read <= alignment_quality:
		# use none of rev and all of fwd
		fwd_use_idx = np.arange(len(fwd_seq))
		rev_use_idx = np.array([])
	
	elif fwd_align_qual_read <= alignment_quality and rev_align_qual_read > alignment_quality:
		# use all of rev and none of fwd
		fwd_use_idx = np.array([])
		rev_use_idx = np.arange(len(rev_seq))
	
	# handle fwd region
	fwd_aligned_pairs = fwd_read.get_aligned_pairs(True)
	fwd_region = [pair for pair in fwd_aligned_pairs if pair[0] in fwd_use_idx]
	for qpos, refpos in fwd_region:
		if refpos is not None and fwd_quality[qpos] > base_qual:
			if fwd_seq[qpos] == "A":
				qualA_fw[refpos] += fwd_quality[qpos]
				countsA_fw[refpos] += 1
			elif fwd_seq[qpos] == "C":
				qualC_fw[refpos] += fwd_quality[qpos]
				countsC_fw[refpos] += 1
			elif fwd_seq[qpos] == "G":
				qualG_fw[refpos] += fwd_quality[qpos]
				countsG_fw[refpos] += 1
			elif fwd_seq[qpos] == "T":
				qualT_fw[refpos] += fwd_quality[qpos]
				countsT_fw[refpos] += 1
	
	# handle rev region
	rev_aligned_pairs = rev_read.get_aligned_pairs(True)
	rev_region = [pair for pair in rev_aligned_pairs if pair[0] in rev_use_idx]
	for qpos, refpos in rev_region:
		if refpos is not None and rev_quality[qpos] > base_qual:
			if rev_seq[qpos] == "A":
				qualA_rev[refpos] += rev_quality[qpos]
				countsA_rev[refpos] += 1
			elif rev_seq[qpos] == "C":
				qualC_rev[refpos] += rev_quality[qpos]
				countsC_rev[refpos] += 1
			elif rev_seq[qpos] == "G":
				qualG_rev[refpos] += rev_quality[qpos]
				countsG_rev[refpos] += 1
			elif rev_seq[qpos] == "T":
				qualT_rev[refpos] += rev_quality[qpos]
				countsT_rev[refpos] += 1

meanQualA_fw = [round(x/y,1) for x, y in zip(qualA_fw, countsA_fw)]
meanQualC_fw = [round(x/y,1) for x, y in zip(qualC_fw, countsC_fw)]
meanQualG_fw = [round(x/y,1) for x, y in zip(qualG_fw, countsG_fw)]
meanQualT_fw = [round(x/y,1) for x, y in zip(qualT_fw, countsT_fw)]

countsA_fw = [ int(round(elem)) for elem in countsA_fw ]
countsC_fw = [ int(round(elem)) for elem in countsC_fw ]
countsG_fw = [ int(round(elem)) for elem in countsG_fw ]
countsT_fw = [ int(round(elem)) for elem in countsT_fw ]

meanQualA_rev = [round(x/y,1) for x, y in zip(qualA_rev, countsA_rev)]
meanQualC_rev = [round(x/y,1) for x, y in zip(qualC_rev, countsC_rev)]
meanQualG_rev = [round(x/y,1) for x, y in zip(qualG_rev, countsG_rev)]
meanQualT_rev = [round(x/y,1) for x, y in zip(qualT_rev, countsT_rev)]

countsA_rev = [ int(round(elem)) for elem in countsA_rev ]
countsC_rev = [ int(round(elem)) for elem in countsC_rev ]
countsG_rev = [ int(round(elem)) for elem in countsG_rev ]
countsT_rev = [ int(round(elem)) for elem in countsT_rev ]

# Allele Counts
bam = pysam.AlignmentFile(bamfile, "rb")

if(emit_base_qualities == "True"):
	writeSparseMatrix4("A", countsA_fw, meanQualA_fw, countsA_rev, meanQualA_rev)
	writeSparseMatrix4("C", countsC_fw, meanQualC_fw, countsC_rev, meanQualC_rev)
	writeSparseMatrix4("G", countsG_fw, meanQualG_fw, countsG_rev, meanQualG_rev)
	writeSparseMatrix4("T", countsT_fw, meanQualT_fw, countsT_rev, meanQualT_rev)
else:
	writeSparseMatrix2("A", countsA_fw, countsA_rev)
	writeSparseMatrix2("C", countsC_fw, countsC_rev)
	writeSparseMatrix2("G", countsG_fw, countsG_rev)
	writeSparseMatrix2("T", countsT_fw, countsT_rev)

zipped_list = zip(list(countsA_fw),list(countsC_fw),list(countsG_fw),list(countsT_fw), list(countsA_rev),list(countsC_rev),list(countsG_rev),list(countsT_rev))
sums = [sum(item) for item in zipped_list]
writeSparseMatrix("coverage", sums)


import click
import os
import os.path
import sys
import shutil
import random
import string
import itertools
import time
import pysam
import glob

from pkg_resources import get_distribution
from subprocess import call, check_call
from mgatk.mgatkHelp import *
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import SingleQuotedScalarString as sqs
from multiprocessing import Pool

@click.command()
@click.version_option()
@click.option('--input', '-i', default = "", required=True, help='Input; either directory of singular .bam file; see wiki')
@click.option('--output', '-o', default="mgatk_out", help='Output directory for analysis.')
@click.option('--name', '-n', default="mgatk_del",  help='Prefix for project name')

@click.option('--mito-chromosome', '-mc', default = "chrM", required=True, help='Mitochondria chromosome name; see wiki')
@click.option('--ncores', '-c', default = "detect", help='Number of cores to run the main job in parallel.')

@click.option('--cluster', default = "",  help='Message to send to Snakemake to execute jobs on cluster interface; see wiki.')
@click.option('--jobs', default = "0",  help='Max number of jobs to be running concurrently on the cluster interface.')

@click.option('--left-coordinates', '-lc', default = "1", help='Comma separated values for right coordinate of deletions; see wiki')
@click.option('--right-coordinates', '-rc', default = "1000", help='Comma separated values for right coordinate of deletions; see wiki')

@click.option('--read-length', '-rl', default = "72", help='Expected length of a single read from the .bam file')
@click.option('--window-outer', '-wo', default = "9", help='Number of bases from the start of each read mates ("outer" part of read) ignored for estimating heteroplasmy.')
@click.option('--window-inner', '-wi', default = "24", help='Number of bases near the insert of the read mates ("outer" part of read) ignored for estimating heteroplasmy.')

@click.option('--keep-temp-files', '-z', is_flag=True, help='Keep all intermediate files.')
@click.option('--snake-stdout', '-so', is_flag=True, help='Write snakemake log to sdout rather than a file.')

def main(input, output, name, mito_chromosome, ncores,
	cluster, jobs, 
	left_coordinates, right_coordinates,
	read_length, window_outer, window_inner,
	keep_temp_files, snake_stdout):
	
	"""
	mgatk-del: quantify deletion heteroplasmy in mtDNA. \n
	"""
	
	script_dir = os.path.dirname(os.path.realpath(__file__))
	cwd = os.getcwd()
	__version__ = get_distribution('mgatk').version
	click.echo(gettime() + "mgatk-del v%s" % __version__)
	
	# Determine cores
	if(ncores == "detect"):
		ncores = str(available_cpu_count())
	else:
		ncores = str(ncores)
	
	# Verify R dependencies
	check_software_exists("R")
	check_R_packages(["dplyr"])
	mito_chr = mito_chromosome
	
	# -------------------------------
	# Determine samples for analysis
	# -------------------------------
	bams = []
	#bams = os.popen('ls ' + input + '/*.bam').read().strip().split("\n")
	bams = glob.glob(input + '/*.bam')

	if bams[0] == '':
		sys.exit('ERROR: Could not import any samples from the user specification; check --input parameter; QUITTING')

	samples = []
	samplebams = []

	# Loop over bam files
	for bam in bams:
		base=os.path.basename(bam)
		basename=os.path.splitext(base)[0]
		samples.append(basename)
		samplebams.append(bam)
		
	# parallel process to ensure that we have .bai files for each bam
	pool = Pool(processes=int(ncores))
	pm = pool.map(verify_bai, samplebams)
	pool.close()
	
	samples_fail = []
	for i in range(len(samples)):
		sample = samples[i]
		bam = samplebams[i]
		mito_length = -9
		if( not verify_sample_mitobam(bam, mito_chr, mito_length)):
			samples_fail.append(sample)
			
	if(len(samples_fail) > 0):
		click.echo(gettime() + "NOTE: the samples below either have 0 mtDNA reads at the specified chromosome or are mapped to an incorrectly specified reference mitochondrial genome")
		click.echo(gettime() + "Will remove samples from processing:" )
		rmidx = findIdx(samples, samples_fail)
		for index in sorted(rmidx, reverse=True):
			print("REMOVED: ", samples[index])
			del samples[index]
			del samplebams[index]
		
	if not len(samples) > 0:
		sys.exit('ERROR: Could not import any samples from the user specification. \nERROR: check flags, logs, and input configuration (including reference mitochondrial genome); \nQUITTING')
	
	# Make all of the output folders if necessary
	of = output; tf = of + "/temp"; logs = of + "/logs"
	folders = [of, tf, logs, of + "/.internal", tf + "/del",
		 of + "/.internal/parseltongue", of + "/.internal/samples", of + "/final"]

	mkfolderout = [make_folder(x) for x in folders]
		

	# Create internal README files 
	if not os.path.exists(of + "/.internal/README"):
		with open(of + "/.internal/README" , 'w') as outfile:
			outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")
	if not os.path.exists(of + "/.internal/parseltongue/README"):	
		with open(of + "/.internal/parseltongue/README" , 'w') as outfile:
			outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")
	if not os.path.exists(of + "/.internal/samples/README"):
		with open(of + "/.internal" + "/samples" + "/README" , 'w') as outfile:
			outfile.write("This folder creates samples to be interpreted by Snakemake; don't modify it.\n\n")
	
	nsamplesNote = "mgatk-del will process " + str(len(samples)) + " samples"
	click.echo(gettime() + "%s" % nsamplesNote)
	
	# Set up sample bam plain text file
	for i in range(len(samples)):
		with open(of + "/.internal/samples/" + samples[i] + ".bam.txt" , 'w') as outfile:
			outfile.write(samplebams[i])
	
	click.echo(gettime() + "Genotyping samples with "+ncores+" threads.")
	
	# add sqs to get .yaml to play friendly https://stackoverflow.com/questions/39262556/preserve-quotes-and-also-add-data-with-quotes-in-ruamel
	
	# Note: during variable name updating, we changed
	# near == outer; far == inner
	dict1 = {'input_directory' : sqs(input), 'output_directory' : sqs(output), 'script_dir' : sqs(script_dir),
		'name' : sqs(name),
		'left_coordinates' : sqs(left_coordinates), 'right_coordinates' : sqs(right_coordinates),
		'read_length' : sqs(read_length),
		'window_near' : sqs(window_outer), 'window_far' : sqs(window_inner)}
	
	# Potentially submit jobs to cluster
	snakeclust = ""
	njobs = int(jobs)
	if(njobs > 0 and cluster != ""):
		snakeclust = " --jobs " + jobs + " --cluster '" + cluster + "' "
		click.echo(gettime() + "Recognized flags to process jobs on a computing cluster.")
			
	y_s = of + "/.internal/parseltongue/snake.dels.yaml"
	with open(y_s, 'w') as yaml_file:
		yaml=YAML()
		yaml.default_flow_style = False
		yaml.dump(dict1, yaml_file)
	
	cp_call = "cp " + y_s +  " " + logs + "/" + name + ".parameters_del.txt"
	os.system(cp_call)
	
	# Execute snakemake
	snake_stats = logs + "/" + name + ".snakemake_del.stats"
	snake_log = logs + "/" + name + ".snakemake_del.log"
	
	snake_log_out = ""
	if not snake_stdout:
		snake_log_out = ' &>' + snake_log
		
	snakecmd_del = 'snakemake'+snakeclust+' --snakefile ' + script_dir + '/singles_del/Snakefile.delSingles --cores '+ncores+' --config cfp="'  + y_s + '" --stats '+snake_stats + snake_log_out
	os.system(snakecmd_del)
	click.echo(gettime() + "mgatk-del successfully processed the supplied .bam files")
	click.echo(gettime() + "Successfully created final output files")
	
	if keep_temp_files:
		click.echo(gettime() + "Retaining temporary files.")
	else:
		byefolder = of
		shutil.rmtree(byefolder + "/.internal")
		shutil.rmtree(byefolder + "/temp")
		click.echo(gettime() + "Intermediate files successfully removed.")
	
